@@ -36,6 +36,52 @@
 
 #include "cairoint.h"
 
+const cairo_image_surface_t _cairo_image_surface_nil_invalid_format = {
+    {
+	&cairo_image_surface_backend,	/* backend */
+	CAIRO_SURFACE_TYPE_IMAGE,
+	CAIRO_CONTENT_COLOR,
+	CAIRO_REF_COUNT_INVALID,	/* ref_count */
+	CAIRO_STATUS_INVALID_FORMAT,	/* status */
+	FALSE,				/* finished */
+	{ 0,	/* size */
+	  0,	/* num_elements */
+	  0,	/* element_size */
+	  NULL,	/* elements */
+	},				/* user_data */
+	{ 1.0, 0.0,
+	  0.0, 1.0,
+	  0.0, 0.0
+	},				/* device_transform */
+	{ 1.0, 0.0,
+	  0.0, 1.0,
+	  0.0, 0.0
+	},				/* device_transform_inverse */
+	0.0,				/* x_fallback_resolution */
+	0.0,				/* y_fallback_resolution */
+	NULL,				/* clip */
+	0,				/* next_clip_serial */
+	0,				/* current_clip_serial */
+	FALSE,				/* is_snapshot */
+	FALSE,				/* has_font_options */
+	{ CAIRO_ANTIALIAS_DEFAULT,
+	  CAIRO_SUBPIXEL_ORDER_DEFAULT,
+	  CAIRO_HINT_STYLE_DEFAULT,
+	  CAIRO_HINT_METRICS_DEFAULT
+	}				/* font_options */
+    },					/* base */
+    CAIRO_FORMAT_ARGB32,		/* format */
+    NULL,				/* data */
+    FALSE,				/* owns_data */
+    FALSE,				/* has_clip */
+    0,					/* width */
+    0,					/* height */
+    0,					/* stride */
+    0,					/* depth */
+    NULL				/* pixman_image */
+};
+
+
 static int
 _cairo_format_bpp (cairo_format_t format)
 {
@@ -44,8 +90,6 @@ _cairo_format_bpp (cairo_format_t format)
 	return 1;
     case CAIRO_FORMAT_A8:
 	return 8;
-    case CAIRO_FORMAT_RGB16_565:
-	return 16;
     case CAIRO_FORMAT_RGB24:
     case CAIRO_FORMAT_ARGB32:
 	return 32;
@@ -142,13 +186,13 @@ _cairo_format_from_pixman_format (pixman_format_t *pixman_format)
     }
 
     fprintf (stderr,
-	     "Error: Cairo does not yet support the requested image format:\n"
+	     "Error: Cairo " PACKAGE_VERSION " does not yet support the requested image format:\n"
 	     "\tDepth: %d\n"
 	     "\tAlpha mask: 0x%08x\n"
 	     "\tRed   mask: 0x%08x\n"
 	     "\tGreen mask: 0x%08x\n"
 	     "\tBlue  mask: 0x%08x\n"
-	     "Please file an enhacement request (quoting the above) at:\n"
+	     "Please file an enhancement request (quoting the above) at:\n"
 	     PACKAGE_BUGREPORT "\n",
 	     bpp, am, rm, gm, bm);
 
@@ -211,9 +255,6 @@ _create_pixman_format (cairo_format_t format)
     case CAIRO_FORMAT_A8:
 	return pixman_format_create (PIXMAN_FORMAT_NAME_A8);
 	break;
-    case CAIRO_FORMAT_RGB16_565:
-	return pixman_format_create (PIXMAN_FORMAT_NAME_RGB16_565);
-	break;
     case CAIRO_FORMAT_RGB24:
 	return pixman_format_create (PIXMAN_FORMAT_NAME_RGB24);
 	break;
@@ -253,8 +294,10 @@ cairo_image_surface_create (cairo_format_t	format,
     pixman_format_t *pixman_format;
     pixman_image_t *pixman_image;
 
-    if (! CAIRO_FORMAT_VALID (format))
-	return (cairo_surface_t*) &_cairo_surface_nil;
+    if (! CAIRO_FORMAT_VALID (format)) {
+	_cairo_error (CAIRO_STATUS_INVALID_FORMAT);
+	return (cairo_surface_t*) &_cairo_image_surface_nil_invalid_format;
+    }
 
     pixman_format = _create_pixman_format (format);
     if (pixman_format == NULL) {
@@ -275,6 +318,7 @@ cairo_image_surface_create (cairo_format_t	format,
 
     return surface;
 }
+slim_hidden_def (cairo_image_surface_create);
 
 cairo_surface_t *
 _cairo_image_surface_create_with_content (cairo_content_t	content,
@@ -303,8 +347,8 @@ _cairo_image_surface_create_with_content (cairo_content_t	content,
  * Creates an image surface for the provided pixel data. The output
  * buffer must be kept around until the #cairo_surface_t is destroyed
  * or cairo_surface_finish() is called on the surface.  The initial
- * contents of @buffer will be used as the inital image contents; you
- * must explicitely clear the buffer, using, for example,
+ * contents of @buffer will be used as the initial image contents; you
+ * must explicitly clear the buffer, using, for example,
  * cairo_rectangle() and cairo_fill() if you want it cleared.
  *
  * Return value: a pointer to the newly created surface. The caller
@@ -354,6 +398,7 @@ cairo_image_surface_create_for_data (unsigned char     *data,
 
     return surface;
 }
+slim_hidden_def (cairo_image_surface_create_for_data);
 
 cairo_surface_t *
 _cairo_image_surface_create_for_data_with_content (unsigned char	*data,
@@ -440,6 +485,7 @@ cairo_image_surface_get_width (cairo_surface_t *surface)
 
     return image_surface->width;
 }
+slim_hidden_def (cairo_image_surface_get_width);
 
 /**
  * cairo_image_surface_get_height:
@@ -461,6 +507,7 @@ cairo_image_surface_get_height (cairo_surface_t *surface)
 
     return image_surface->height;
 }
+slim_hidden_def (cairo_image_surface_get_height);
 
 /**
  * cairo_image_surface_get_stride:
@@ -618,6 +665,10 @@ _cairo_image_surface_release_dest_image (void                    *abstract_surfa
 static cairo_status_t
 _cairo_image_surface_clone_similar (void		*abstract_surface,
 				    cairo_surface_t	*src,
+				    int                  src_x,
+				    int                  src_y,
+				    int                  width,
+				    int                  height,
 				    cairo_surface_t    **clone_out)
 {
     cairo_image_surface_t *surface = abstract_surface;
@@ -1050,3 +1101,29 @@ const cairo_surface_backend_t cairo_image_surface_backend = {
     _cairo_image_surface_get_extents,
     NULL /* old_show_glyphs */
 };
+
+/* A convenience function for when one needs to coerce an image
+ * surface to an alternate format. */
+cairo_image_surface_t *
+_cairo_image_surface_clone (cairo_image_surface_t	*surface,
+			    cairo_format_t		 format)
+{
+    cairo_image_surface_t *clone;
+    cairo_t *cr;
+    double x, y;
+
+    clone = (cairo_image_surface_t *)
+	cairo_image_surface_create (format,
+				    surface->width, surface->height);
+
+    cr = cairo_create (&clone->base);
+    cairo_surface_get_device_offset (&surface->base, &x, &y);
+    cairo_set_source_surface (cr, &surface->base, x, y);
+    cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
+    cairo_paint (cr);
+    cairo_destroy (cr);
+
+    cairo_surface_set_device_offset (&clone->base, x, y);
+
+    return clone;
+}
