@@ -226,13 +226,15 @@ void
 _cairo_scaled_font_map_destroy (void)
 {
     int i;
-    cairo_scaled_font_map_t *font_map = cairo_scaled_font_map;
+    cairo_scaled_font_map_t *font_map;
     cairo_scaled_font_t *scaled_font;
 
-    if (font_map == NULL)
-	return;
+    CAIRO_MUTEX_LOCK (cairo_scaled_font_map_mutex);
 
-    CAIRO_MUTEX_UNLOCK (cairo_scaled_font_map_mutex);
+    font_map = cairo_scaled_font_map;
+    if (font_map == NULL) {
+        goto CLEANUP_MUTEX_LOCK;
+    }
 
     for (i = 0; i < font_map->num_holdovers; i++) {
 	scaled_font = font_map->holdovers[i];
@@ -250,6 +252,9 @@ _cairo_scaled_font_map_destroy (void)
 
     free (cairo_scaled_font_map);
     cairo_scaled_font_map = NULL;
+
+ CLEANUP_MUTEX_LOCK:
+    CAIRO_MUTEX_UNLOCK (cairo_scaled_font_map_mutex);
 }
 
 /* Fowler / Noll / Vo (FNV) Hash (http://www.isthe.com/chongo/tech/comp/fnv/)
@@ -321,7 +326,7 @@ _cairo_scaled_font_keys_equal (const void *abstract_key_a, const void *abstract_
 
 /* XXX: This 256 number is arbitary---we've never done any measurement
  * of this. In fact, having a per-font glyph caches each managed
- * separately is probably not waht we want anyway. Would probably be
+ * separately is probably not what we want anyway. Would probably be
  * much better to have a single cache for glyphs with random
  * replacement across all glyphs of all fonts. */
 static int max_glyphs_cached_per_font = 256;
@@ -493,6 +498,7 @@ UNWIND_FONT_MAP_LOCK:
 UNWIND:
     return NULL;
 }
+slim_hidden_def (cairo_scaled_font_create);
 
 /**
  * cairo_scaled_font_reference:
@@ -551,6 +557,7 @@ cairo_scaled_font_reference (cairo_scaled_font_t *scaled_font)
 
     return scaled_font;
 }
+slim_hidden_def (cairo_scaled_font_reference);
 
 /**
  * cairo_scaled_font_destroy:
@@ -611,6 +618,7 @@ cairo_scaled_font_destroy (cairo_scaled_font_t *scaled_font)
     }
     _cairo_scaled_font_map_unlock ();
 }
+slim_hidden_def (cairo_scaled_font_destroy);
 
 /* Public font API follows. */
 
@@ -627,6 +635,7 @@ cairo_scaled_font_extents (cairo_scaled_font_t  *scaled_font,
 {
     *extents = scaled_font->extents;
 }
+slim_hidden_def (cairo_scaled_font_extents);
 
 /**
  * cairo_scaled_font_text_extents:
@@ -689,31 +698,20 @@ cairo_scaled_font_text_extents (cairo_scaled_font_t   *scaled_font,
  **/
 void
 cairo_scaled_font_glyph_extents (cairo_scaled_font_t   *scaled_font,
-				 cairo_glyph_t         *glyphs,
+				 const cairo_glyph_t   *glyphs,
 				 int                    num_glyphs,
 				 cairo_text_extents_t  *extents)
 {
     cairo_status_t status = CAIRO_STATUS_SUCCESS;
     int i;
     double min_x = 0.0, min_y = 0.0, max_x = 0.0, max_y = 0.0;
-    double x_pos = 0.0, y_pos = 0.0;
+    cairo_bool_t visible = FALSE;
+    cairo_scaled_glyph_t *scaled_glyph = NULL;
 
     if (scaled_font->status)
 	return;
 
-    if (!num_glyphs) {
-	extents->x_bearing = 0.0;
-	extents->y_bearing = 0.0;
-	extents->width = 0.0;
-	extents->height = 0.0;
-	extents->x_advance = 0.0;
-	extents->y_advance = 0.0;
-
-	return;
-    }
-
     for (i = 0; i < num_glyphs; i++) {
-	cairo_scaled_glyph_t	*scaled_glyph;
 	double			left, top, right, bottom;
 
 	status = _cairo_scaled_glyph_lookup (scaled_font,
@@ -725,12 +723,17 @@ cairo_scaled_font_glyph_extents (cairo_scaled_font_t   *scaled_font,
 	    return;
 	}
 
+	/* "Ink" extents should skip "invisible" glyphs */
+	if (scaled_glyph->metrics.width == 0 && scaled_glyph->metrics.height == 0)
+	    continue;
+
 	left = scaled_glyph->metrics.x_bearing + glyphs[i].x;
 	right = left + scaled_glyph->metrics.width;
 	top = scaled_glyph->metrics.y_bearing + glyphs[i].y;
 	bottom = top + scaled_glyph->metrics.height;
 
-	if (i == 0) {
+	if (!visible) {
+	    visible = TRUE;
 	    min_x = left;
 	    max_x = right;
 	    min_y = top;
@@ -741,17 +744,38 @@ cairo_scaled_font_glyph_extents (cairo_scaled_font_t   *scaled_font,
 	    if (top < min_y) min_y = top;
 	    if (bottom > max_y) max_y = bottom;
 	}
-	x_pos = glyphs[i].x + scaled_glyph->metrics.x_advance;
-	y_pos = glyphs[i].y + scaled_glyph->metrics.y_advance;
     }
 
-    extents->x_bearing = min_x - glyphs[0].x;
-    extents->y_bearing = min_y - glyphs[0].y;
-    extents->width = max_x - min_x;
-    extents->height = max_y - min_y;
-    extents->x_advance = x_pos - glyphs[0].x;
-    extents->y_advance = y_pos - glyphs[0].y;
+    if (visible) {
+	extents->x_bearing = min_x - glyphs[0].x;
+	extents->y_bearing = min_y - glyphs[0].y;
+	extents->width = max_x - min_x;
+	extents->height = max_y - min_y;
+    } else {
+	extents->x_bearing = 0.0;
+	extents->y_bearing = 0.0;
+	extents->width = 0.0;
+	extents->height = 0.0;
+    }
+
+    if (num_glyphs) {
+        double x0, y0, x1, y1;
+
+	x0 = glyphs[0].x;
+	y0 = glyphs[0].y;
+
+	/* scaled_glyphs contains the glyph for num_glyphs - 1 already. */
+	x1 = glyphs[num_glyphs - 1].x + scaled_glyph->metrics.x_advance;
+	y1 = glyphs[num_glyphs - 1].y + scaled_glyph->metrics.y_advance;
+
+	extents->x_advance = x1 - x0;
+	extents->y_advance = y1 - y0;
+    } else {
+	extents->x_advance = 0.0;
+	extents->y_advance = 0.0;
+    }
 }
+slim_hidden_def (cairo_scaled_font_glyph_extents);
 
 cairo_status_t
 _cairo_scaled_font_text_to_glyphs (cairo_scaled_font_t *scaled_font,
@@ -766,6 +790,12 @@ _cairo_scaled_font_text_to_glyphs (cairo_scaled_font_t *scaled_font,
     cairo_status_t status = CAIRO_STATUS_SUCCESS;
     cairo_scaled_glyph_t *scaled_glyph;
 
+    if (utf8[0] == '\0') {
+	*num_glyphs = 0;
+	*glyphs = NULL;
+	return CAIRO_STATUS_SUCCESS;
+    }
+
     if (scaled_font->backend->text_to_glyphs) {
 	status = scaled_font->backend->text_to_glyphs (scaled_font,
 						       x, y, utf8,
@@ -824,7 +854,7 @@ _cairo_scaled_font_glyph_device_extents (cairo_scaled_font_t	 *scaled_font,
     cairo_status_t status = CAIRO_STATUS_SUCCESS;
     int i;
     int min_x = INT16_MAX, max_x = INT16_MIN;
-    int	min_y = INT16_MAX, max_y = INT16_MAX;
+    int	min_y = INT16_MAX, max_y = INT16_MIN;
 
     if (scaled_font->status)
 	return scaled_font->status;
@@ -845,8 +875,8 @@ _cairo_scaled_font_glyph_device_extents (cairo_scaled_font_t	 *scaled_font,
 	}
 
 	/* glyph images are snapped to pixel locations */
-	x = (int) floor (glyphs[i].x + 0.5);
-	y = (int) floor (glyphs[i].y + 0.5);
+	x = _cairo_lround (glyphs[i].x);
+	y = _cairo_lround (glyphs[i].y);
 
 	left   = x + _cairo_fixed_integer_floor(scaled_glyph->bbox.p1.x);
 	top    = y + _cairo_fixed_integer_floor (scaled_glyph->bbox.p1.y);
@@ -881,7 +911,7 @@ _cairo_scaled_font_show_glyphs (cairo_scaled_font_t    *scaled_font,
 				int			dest_y,
 				unsigned int		width,
 				unsigned int		height,
-				const cairo_glyph_t    *glyphs,
+				cairo_glyph_t	       *glyphs,
 				int                     num_glyphs)
 {
     cairo_status_t status;
@@ -896,6 +926,9 @@ _cairo_scaled_font_show_glyphs (cairo_scaled_font_t    *scaled_font,
     if (scaled_font->status)
 	return scaled_font->status;
 
+    if (!num_glyphs)
+	return CAIRO_STATUS_SUCCESS;
+
     if (scaled_font->backend->show_glyphs != NULL) {
 	status = scaled_font->backend->show_glyphs (scaled_font,
 						    op, pattern,
@@ -954,12 +987,10 @@ _cairo_scaled_font_show_glyphs (cairo_scaled_font_t    *scaled_font,
 
 	/* round glyph locations to the nearest pixel */
 	/* XXX: FRAGILE: We're ignoring device_transform scaling here. A bug? */
-	x = (int) floor (glyphs[i].x +
-                         glyph_surface->base.device_transform.x0 +
-                         0.5);
-	y = (int) floor (glyphs[i].y +
-                         glyph_surface->base.device_transform.y0 +
-                         0.5);
+	x = _cairo_lround (glyphs[i].x +
+                           glyph_surface->base.device_transform.x0);
+	y = _cairo_lround (glyphs[i].y +
+                           glyph_surface->base.device_transform.y0);
 
 	_cairo_pattern_init_for_surface (&glyph_pattern, &glyph_surface->base);
 
@@ -1052,6 +1083,74 @@ _scaled_glyph_path_close_path (void *abstract_closure)
     return _cairo_path_fixed_close_path (closure->path);
 }
 
+
+/**
+ * _trace_mask_to_path:
+ * @bitmap: An alpha mask (either CAIRO_FORMAT_A1 or _A8)
+ * @path: An initialized path to hold the result
+ *
+ * Given a mask surface, (an alpha image), fill out the provided path
+ * so that when filled it would result in something that approximates
+ * the mask.
+ *
+ * Note: The current tracing code here is extremely primitive. It
+ * operates only on an A1 surface, (converting an A8 surface to A1 if
+ * necessary), and performs the tracing by drawing a little square
+ * around each pixel that is on in the mask. We do not pretend that
+ * this is a high-quality result. But we are leaving it up to someone
+ * who cares enough about getting a better result to implement
+ * something more sophisticated.
+ **/
+static cairo_status_t
+_trace_mask_to_path (cairo_image_surface_t *mask,
+		     cairo_path_fixed_t *path)
+{
+    cairo_image_surface_t *a1_mask;
+    unsigned char *row, *byte_ptr, byte;
+    int rows, cols, bytes_per_row;
+    int x, y, bit;
+    double xoff, yoff;
+
+    if (mask->format == CAIRO_FORMAT_A1)
+	a1_mask = mask;
+    else
+	a1_mask = _cairo_image_surface_clone (mask, CAIRO_FORMAT_A1);
+
+    if (cairo_surface_status (&a1_mask->base))
+	return cairo_surface_status (&a1_mask->base);
+
+    cairo_surface_get_device_offset (&mask->base, &xoff, &yoff);
+
+    bytes_per_row = (a1_mask->width + 7) / 8;
+    for (y = 0, row = a1_mask->data, rows = a1_mask->height; rows; row += a1_mask->stride, rows--, y++) {
+	for (x = 0, byte_ptr = row, cols = (a1_mask->width + 7) / 8; cols; byte_ptr++, cols--) {
+	    byte = CAIRO_BITSWAP8_IF_LITTLE_ENDIAN (*byte_ptr);
+	    for (bit = 7; bit >= 0 && x < a1_mask->width; bit--, x++) {
+		if (byte & (1 << bit)) {
+		    _cairo_path_fixed_move_to (path,
+					       _cairo_fixed_from_int (x + xoff),
+					       _cairo_fixed_from_int (y + yoff));
+		    _cairo_path_fixed_rel_line_to (path,
+						   _cairo_fixed_from_int (1),
+						   _cairo_fixed_from_int (0));
+		    _cairo_path_fixed_rel_line_to (path,
+						   _cairo_fixed_from_int (0),
+						   _cairo_fixed_from_int (1));
+		    _cairo_path_fixed_rel_line_to (path,
+						   _cairo_fixed_from_int (-1),
+						   _cairo_fixed_from_int (0));
+		    _cairo_path_fixed_close_path (path);
+		}
+	    }
+	}
+    }
+
+    if (a1_mask != mask)
+	cairo_surface_destroy (&a1_mask->base);
+
+    return CAIRO_STATUS_SUCCESS;
+}
+
 cairo_status_t
 _cairo_scaled_font_glyph_path (cairo_scaled_font_t *scaled_font,
 			       const cairo_glyph_t *glyphs,
@@ -1061,6 +1160,7 @@ _cairo_scaled_font_glyph_path (cairo_scaled_font_t *scaled_font,
     cairo_status_t status;
     int	i;
     cairo_scaled_glyph_path_closure_t closure;
+    cairo_path_fixed_t *glyph_path;
 
     if (scaled_font->status)
 	return scaled_font->status;
@@ -1073,19 +1173,42 @@ _cairo_scaled_font_glyph_path (cairo_scaled_font_t *scaled_font,
 					     glyphs[i].index,
 					     CAIRO_SCALED_GLYPH_INFO_PATH,
 					     &scaled_glyph);
-	if (status)
+	if (status == CAIRO_STATUS_SUCCESS)
+	    glyph_path = scaled_glyph->path;
+	else if (status != CAIRO_INT_STATUS_UNSUPPORTED)
 	    return status;
 
+	/* If the font is incapable of providing a path, then we'll
+	 * have to trace our own from a surface. */
+	if (status == CAIRO_INT_STATUS_UNSUPPORTED) {
+	    status = _cairo_scaled_glyph_lookup (scaled_font,
+						 glyphs[i].index,
+						 CAIRO_SCALED_GLYPH_INFO_SURFACE,
+						 &scaled_glyph);
+
+	    glyph_path = _cairo_path_fixed_create ();
+	    if (glyph_path == NULL)
+		return CAIRO_STATUS_NO_MEMORY;
+
+	    status = _trace_mask_to_path (scaled_glyph->surface, glyph_path);
+	    if (status) {
+		_cairo_path_fixed_destroy (glyph_path);
+		return status;
+	    }
+	}
+
 	closure.offset.x = _cairo_fixed_from_double (glyphs[i].x);
 	closure.offset.y = _cairo_fixed_from_double (glyphs[i].y);
 
-	status = _cairo_path_fixed_interpret (scaled_glyph->path,
+	status = _cairo_path_fixed_interpret (glyph_path,
 					      CAIRO_DIRECTION_FORWARD,
 					      _scaled_glyph_path_move_to,
 					      _scaled_glyph_path_line_to,
 					      _scaled_glyph_path_curve_to,
 					      _scaled_glyph_path_close_path,
 					      &closure);
+	if (glyph_path != scaled_glyph->path)
+	    _cairo_path_fixed_destroy (glyph_path);
     }
 
     return CAIRO_STATUS_SUCCESS;
@@ -1111,6 +1234,7 @@ _cairo_scaled_glyph_set_metrics (cairo_scaled_glyph_t *scaled_glyph,
     double hm, wm;
     double min_user_x = 0.0, max_user_x = 0.0, min_user_y = 0.0, max_user_y = 0.0;
     double min_device_x = 0.0, max_device_x = 0.0, min_device_y = 0.0, max_device_y = 0.0;
+    double device_x_advance, device_y_advance;
 
     for (hm = 0.0; hm <= 1.0; hm += 1.0)
 	for (wm = 0.0; wm <= 1.0; wm += 1.0) {
@@ -1159,10 +1283,19 @@ _cairo_scaled_glyph_set_metrics (cairo_scaled_glyph_t *scaled_glyph,
 				     &scaled_glyph->metrics.x_advance,
 				     &scaled_glyph->metrics.y_advance);
 
+    device_x_advance = fs_metrics->x_advance;
+    device_y_advance = fs_metrics->y_advance;
+    cairo_matrix_transform_distance (&scaled_font->scale,
+				     &device_x_advance,
+				     &device_y_advance);
+
     scaled_glyph->bbox.p1.x = _cairo_fixed_from_double (min_device_x);
     scaled_glyph->bbox.p1.y = _cairo_fixed_from_double (min_device_y);
     scaled_glyph->bbox.p2.x = _cairo_fixed_from_double (max_device_x);
     scaled_glyph->bbox.p2.y = _cairo_fixed_from_double (max_device_y);
+
+    scaled_glyph->x_advance = _cairo_lround (device_x_advance);
+    scaled_glyph->y_advance = _cairo_lround (device_y_advance);
 }
 
 void
@@ -1316,6 +1449,7 @@ cairo_scaled_font_get_font_face (cairo_scaled_font_t *scaled_font)
 
     return scaled_font->font_face;
 }
+slim_hidden_def (cairo_scaled_font_get_font_face);
 
 /**
  * cairo_scaled_font_get_font_matrix:
@@ -1338,6 +1472,7 @@ cairo_scaled_font_get_font_matrix (cairo_scaled_font_t	*scaled_font,
 
     *font_matrix = scaled_font->font_matrix;
 }
+slim_hidden_def (cairo_scaled_font_get_font_matrix);
 
 /**
  * cairo_scaled_font_get_ctm:
@@ -1359,6 +1494,7 @@ cairo_scaled_font_get_ctm (cairo_scaled_font_t	*scaled_font,
 
     *ctm = scaled_font->ctm;
 }
+slim_hidden_def (cairo_scaled_font_get_ctm);
 
 /**
  * cairo_scaled_font_get_font_options:
@@ -1381,3 +1517,4 @@ cairo_scaled_font_get_font_options (cairo_scaled_font_t		*scaled_font,
 
     _cairo_font_options_init_copy (options, &scaled_font->options);
 }
+slim_hidden_def (cairo_scaled_font_get_font_options);
