@@ -50,6 +50,8 @@
 #define TT_PRIM_CSPLINE 3
 #endif
 
+#define OPENTYPE_CFF_TAG 0x20464643
+
 const cairo_scaled_font_backend_t cairo_win32_scaled_font_backend;
 
 typedef struct {
@@ -97,6 +99,8 @@ typedef struct {
     HFONT scaled_hfont;
     HFONT unscaled_hfont;
 
+    cairo_bool_t glyph_indexing;
+
     cairo_bool_t delete_scaled_hfont;
 } cairo_win32_scaled_font_t;
 
@@ -147,7 +151,8 @@ _compute_transform (cairo_win32_scaled_font_t *scaled_font,
 	    scaled_font->y_scale = - scaled_font->y_scale;
 
 	scaled_font->logical_scale = WIN32_FONT_LOGICAL_SCALE * scaled_font->y_scale;
-	scaled_font->logical_size = WIN32_FONT_LOGICAL_SCALE * floor (scaled_font->y_scale + 0.5);
+	scaled_font->logical_size = WIN32_FONT_LOGICAL_SCALE *
+                                    _cairo_lround (scaled_font->y_scale);
     }
 
     /* The font matrix has x and y "scale" components which we extract and
@@ -161,7 +166,8 @@ _compute_transform (cairo_win32_scaled_font_t *scaled_font,
 					     &scaled_font->x_scale, &scaled_font->y_scale,
 					     TRUE);	/* XXX: Handle vertical text */
 
-	scaled_font->logical_size = floor (WIN32_FONT_LOGICAL_SCALE * scaled_font->y_scale + 0.5);
+	scaled_font->logical_size = _cairo_lround (WIN32_FONT_LOGICAL_SCALE *
+                                                   scaled_font->y_scale);
 	scaled_font->logical_scale = WIN32_FONT_LOGICAL_SCALE * scaled_font->y_scale;
     }
 
@@ -233,6 +239,8 @@ _win32_scaled_font_create (LOGFONTW                   *logfont,
     cairo_matrix_t scale;
     cairo_status_t status;
 
+    _cairo_win32_initialize ();
+
     f = malloc (sizeof(cairo_win32_scaled_font_t));
     if (f == NULL)
 	return NULL;
@@ -464,6 +472,8 @@ _cairo_win32_scaled_font_create_toy (cairo_toy_font_face_t *toy_face,
     int face_name_len;
     cairo_status_t status;
 
+    _cairo_win32_initialize ();
+
     status = _cairo_utf8_to_utf16 (toy_face->family, -1,
 				   &face_name, &face_name_len);
     if (status)
@@ -720,6 +730,12 @@ _cairo_win32_scaled_font_set_metrics (cairo_win32_scaled_font_t *scaled_font)
 
     }
 
+    if ((metrics.tmPitchAndFamily & TMPF_TRUETYPE) ||
+        (GetFontData (hdc, OPENTYPE_CFF_TAG, 0, NULL, 0) != GDI_ERROR))
+        scaled_font->glyph_indexing = TRUE;
+    else
+        scaled_font->glyph_indexing = FALSE;
+
     _cairo_scaled_font_set_metrics (&scaled_font->base, &extents);
 
     return CAIRO_STATUS_SUCCESS;
@@ -734,12 +750,18 @@ _cairo_win32_scaled_font_init_glyph_metrics (cairo_win32_scaled_font_t *scaled_f
     cairo_status_t status;
     cairo_text_extents_t extents;
     HDC hdc;
+    UINT glyph_index_option;
 
     hdc = _get_global_font_dc ();
     if (!hdc)
 	return CAIRO_STATUS_NO_MEMORY;
 
-    if (scaled_font->preserve_axes) {
+    if (scaled_font->glyph_indexing)
+        glyph_index_option = GGO_GLYPH_INDEX;
+    else
+        glyph_index_option = 0;
+
+    if (scaled_font->preserve_axes && scaled_font->base.options.hint_style != CAIRO_HINT_METRICS_OFF) {
 	/* If we aren't rotating / skewing the axes, then we get the metrics
 	 * from the GDI in device space and convert to font space.
 	 */
@@ -747,7 +769,7 @@ _cairo_win32_scaled_font_init_glyph_metrics (cairo_win32_scaled_font_t *scaled_f
 	if (status)
 	    return status;
 	if (GetGlyphOutlineW (hdc, _cairo_scaled_glyph_index (scaled_glyph),
-			      GGO_METRICS | GGO_GLYPH_INDEX,
+			      GGO_METRICS | glyph_index_option,
 			      &metrics, 0, NULL, &matrix) == GDI_ERROR) {
 	  status = _cairo_win32_print_gdi_error ("_cairo_win32_scaled_font_init_glyph_metrics:GetGlyphOutlineW");
 	  memset (&metrics, 0, sizeof (GLYPHMETRICS));
@@ -786,7 +808,7 @@ _cairo_win32_scaled_font_init_glyph_metrics (cairo_win32_scaled_font_t *scaled_f
 	 */
 	status = _cairo_win32_scaled_font_select_unscaled_font (&scaled_font->base, hdc);
 	if (GetGlyphOutlineW (hdc, _cairo_scaled_glyph_index (scaled_glyph),
-			      GGO_METRICS | GGO_GLYPH_INDEX,
+			      GGO_METRICS | glyph_index_option,
 			      &metrics, 0, NULL, &matrix) == GDI_ERROR) {
 	  status = _cairo_win32_print_gdi_error ("_cairo_win32_scaled_font_init_glyph_metrics:GetGlyphOutlineW");
 	  memset (&metrics, 0, sizeof (GLYPHMETRICS));
@@ -829,6 +851,7 @@ _cairo_win32_scaled_font_glyph_bbox (void		 *abstract_font,
 	GLYPHMETRICS metrics;
 	cairo_status_t status;
 	int i;
+        UINT glyph_index_option;
 
 	if (!hdc)
 	    return CAIRO_STATUS_NO_MEMORY;
@@ -837,11 +860,16 @@ _cairo_win32_scaled_font_glyph_bbox (void		 *abstract_font,
 	if (status)
 	    return status;
 
-	for (i = 0; i < num_glyphs; i++) {
-	    int x = floor (0.5 + glyphs[i].x);
-	    int y = floor (0.5 + glyphs[i].y);
+        if (scaled_font->glyph_indexing)
+            glyph_index_option = GGO_GLYPH_INDEX;
+        else
+            glyph_index_option = 0;
 
-	    GetGlyphOutlineW (hdc, glyphs[i].index, GGO_METRICS | GGO_GLYPH_INDEX,
+	for (i = 0; i < num_glyphs; i++) {
+	    int x = _cairo_lround (glyphs[i].x);
+	    int y = _cairo_lround (glyphs[i].y);
+
+	    GetGlyphOutlineW (hdc, glyphs[i].index, GGO_METRICS | glyph_index_option,
 			     &metrics, 0, NULL, &matrix);
 
 	    if (i == 0 || x1 > x + metrics.gmptGlyphOrigin.x)
@@ -897,16 +925,22 @@ _flush_glyphs (cairo_glyph_state_t *state)
     int dx = 0;
     WCHAR * elements;
     int * dx_elements;
+    UINT glyph_index_option;
 
     status = _cairo_array_append (&state->dx, &dx);
     if (status)
 	return status;
 
+    if (state->scaled_font->glyph_indexing)
+        glyph_index_option = ETO_GLYPH_INDEX;
+    else
+        glyph_index_option = 0;
+
     elements = _cairo_array_index (&state->glyphs, 0);
     dx_elements = _cairo_array_index (&state->dx, 0);
     if (!ExtTextOutW (state->hdc,
 		      state->start_x, state->last_y,
-		      ETO_GLYPH_INDEX,
+		      glyph_index_option,
 		      NULL,
 		      elements,
 		      state->glyphs.num_elements,
@@ -934,8 +968,8 @@ _add_glyph (cairo_glyph_state_t *state,
 
     cairo_matrix_transform_point (&state->scaled_font->device_to_logical, &user_x, &user_y);
 
-    logical_x = floor (user_x + 0.5);
-    logical_y = floor (user_y + 0.5);
+    logical_x = _cairo_lround (user_x);
+    logical_y = _cairo_lround (user_y);
 
     if (state->glyphs.num_elements > 0) {
 	int dx;
@@ -1079,7 +1113,7 @@ _compute_a8_mask (cairo_win32_surface_t *mask_surface)
     return &image8->base;
 }
 
-static cairo_status_t
+static cairo_int_status_t
 _cairo_win32_scaled_font_glyph_init (void		       *abstract_font,
 				     cairo_scaled_glyph_t      *scaled_glyph,
 				     cairo_scaled_glyph_info_t  info)
@@ -1117,7 +1151,7 @@ _cairo_win32_scaled_font_show_glyphs (void		       *abstract_font,
 				      int			dest_y,
 				      unsigned int		width,
 				      unsigned int		height,
-				      const cairo_glyph_t      *glyphs,
+				      cairo_glyph_t	       *glyphs,
 				      int                 	num_glyphs)
 {
     cairo_win32_scaled_font_t *scaled_font = abstract_font;
@@ -1239,18 +1273,26 @@ _cairo_win32_scaled_font_load_truetype_table (void	       *abstract_font,
     status = _cairo_win32_scaled_font_select_unscaled_font (&scaled_font->base, hdc);
 
     *length = GetFontData (hdc, tag, offset, buffer, *length);
-    if (*length == GDI_ERROR) {
-        status = _cairo_win32_print_gdi_error ("_cairo_win32_scaled_font_load_truetype_table:GetFontData");
-    }
+    if (*length == GDI_ERROR)
+        status = CAIRO_INT_STATUS_UNSUPPORTED;
+
     _cairo_win32_scaled_font_done_unscaled_font (&scaled_font->base);
 
     return status;
 }
 
-static cairo_fixed_t
-_cairo_fixed_from_FIXED (FIXED f)
+static void
+_cairo_win32_transform_FIXED_to_fixed (cairo_matrix_t *matrix,
+                                       FIXED Fx, FIXED Fy,
+                                       cairo_fixed_t *fx, cairo_fixed_t *fy)
 {
-    return *((cairo_fixed_t *)&f);
+    double x, y;
+
+    x = _cairo_fixed_to_double (*((cairo_fixed_t *)&Fx));
+    y = _cairo_fixed_to_double (*((cairo_fixed_t *)&Fy));
+    cairo_matrix_transform_point (matrix, &x, &y);
+    *fx =  _cairo_fixed_from_double (x);
+    *fy =  _cairo_fixed_from_double (y);
 }
 
 static cairo_status_t
@@ -1264,6 +1306,9 @@ _cairo_win32_scaled_font_init_glyph_path (cairo_win32_scaled_font_t *scaled_font
     DWORD bytesGlyph;
     unsigned char *buffer, *ptr;
     cairo_path_fixed_t *path;
+    cairo_matrix_t transform;
+    cairo_fixed_t x, y;
+    UINT glyph_index_option;
 
     hdc = _get_global_font_dc ();
     if (!hdc)
@@ -1273,12 +1318,24 @@ _cairo_win32_scaled_font_init_glyph_path (cairo_win32_scaled_font_t *scaled_font
     if (!path)
 	return CAIRO_STATUS_NO_MEMORY;
 
-    status = cairo_win32_scaled_font_select_font (&scaled_font->base, hdc);
+    if (scaled_font->base.options.hint_style == CAIRO_HINT_STYLE_NONE) {
+        status = _cairo_win32_scaled_font_select_unscaled_font (&scaled_font->base, hdc);
+        transform = scaled_font->base.scale;
+        cairo_matrix_scale (&transform, 1.0/scaled_font->em_square, 1.0/scaled_font->em_square);
+    } else {
+        status = cairo_win32_scaled_font_select_font (&scaled_font->base, hdc);
+        cairo_matrix_init_identity(&transform);
+    }
     if (status)
         goto CLEANUP_PATH;
 
+    if (scaled_font->glyph_indexing)
+        glyph_index_option = GGO_GLYPH_INDEX;
+    else
+        glyph_index_option = 0;
+
     bytesGlyph = GetGlyphOutlineW (hdc, _cairo_scaled_glyph_index (scaled_glyph),
-				   GGO_NATIVE | GGO_GLYPH_INDEX,
+				   GGO_NATIVE | glyph_index_option,
 				   &metrics, 0, NULL, &matrix);
 
     if (bytesGlyph == GDI_ERROR) {
@@ -1294,7 +1351,7 @@ _cairo_win32_scaled_font_init_glyph_path (cairo_win32_scaled_font_t *scaled_font
     }
 
     if (GetGlyphOutlineW (hdc, _cairo_scaled_glyph_index (scaled_glyph),
-			  GGO_NATIVE | GGO_GLYPH_INDEX,
+			  GGO_NATIVE | glyph_index_option,
 			  &metrics, bytesGlyph, buffer, &matrix) == GDI_ERROR) {
 	status = _cairo_win32_print_gdi_error ("_cairo_win32_scaled_font_glyph_path");
 	free (buffer);
@@ -1307,9 +1364,11 @@ _cairo_win32_scaled_font_init_glyph_path (cairo_win32_scaled_font_t *scaled_font
 
 	ptr += sizeof (TTPOLYGONHEADER);
 
-	_cairo_path_fixed_move_to (path,
-				   _cairo_fixed_from_FIXED (header->pfxStart.x),
-				   _cairo_fixed_from_FIXED (header->pfxStart.y));
+        _cairo_win32_transform_FIXED_to_fixed (&transform,
+                                               header->pfxStart.x,
+                                               header->pfxStart.y,
+                                               &x, &y);
+        _cairo_path_fixed_move_to (path, x, y);
 
 	while (ptr < endPoly) {
 	    TTPOLYCURVE *curve = (TTPOLYCURVE *)ptr;
@@ -1318,26 +1377,36 @@ _cairo_win32_scaled_font_init_glyph_path (cairo_win32_scaled_font_t *scaled_font
 	    switch (curve->wType) {
 	    case TT_PRIM_LINE:
 		for (i = 0; i < curve->cpfx; i++) {
-		    _cairo_path_fixed_line_to (path,
-					       _cairo_fixed_from_FIXED (points[i].x),
-					       _cairo_fixed_from_FIXED (points[i].y));
+                    _cairo_win32_transform_FIXED_to_fixed (&transform,
+                                                           points[i].x,
+                                                           points[i].y,
+                                                           &x, &y);
+		    _cairo_path_fixed_line_to (path, x, y);
 		}
 		break;
 	    case TT_PRIM_QSPLINE:
 		for (i = 0; i < curve->cpfx - 1; i++) {
 		    cairo_fixed_t p1x, p1y, p2x, p2y, cx, cy, c1x, c1y, c2x, c2y;
 		    _cairo_path_fixed_get_current_point (path, &p1x, &p1y);
-		    cx = _cairo_fixed_from_FIXED (points[i].x);
-		    cy = _cairo_fixed_from_FIXED (points[i].y);
+                    _cairo_win32_transform_FIXED_to_fixed (&transform,
+                                                           points[i].x,
+                                                           points[i].y,
+                                                           &cx, &cy);
 
 		    if (i + 1 == curve->cpfx - 1) {
-			p2x = _cairo_fixed_from_FIXED (points[i + 1].x);
-			p2y = _cairo_fixed_from_FIXED (points[i + 1].y);
+                        _cairo_win32_transform_FIXED_to_fixed (&transform,
+                                                               points[i + 1].x,
+                                                               points[i + 1].y,
+                                                               &p2x, &p2y);
 		    } else {
 			/* records with more than one curve use interpolation for
 			   control points, per http://support.microsoft.com/kb/q87115/ */
-			p2x = (cx + _cairo_fixed_from_FIXED (points[i + 1].x)) / 2;
-			p2y = (cy + _cairo_fixed_from_FIXED (points[i + 1].y)) / 2;
+                        _cairo_win32_transform_FIXED_to_fixed (&transform,
+                                                               points[i + 1].x,
+                                                               points[i + 1].y,
+                                                               &x, &y);
+                        p2x = (cx + x) / 2;
+			p2y = (cy + y) / 2;
 		    }
 
 		    c1x = 2 * cx / 3 + p1x / 3;
@@ -1350,13 +1419,20 @@ _cairo_win32_scaled_font_init_glyph_path (cairo_win32_scaled_font_t *scaled_font
 		break;
 	    case TT_PRIM_CSPLINE:
 		for (i = 0; i < curve->cpfx - 2; i += 2) {
-		    _cairo_path_fixed_curve_to (path,
-						_cairo_fixed_from_FIXED (points[i].x),
-						_cairo_fixed_from_FIXED (points[i].y),
-						_cairo_fixed_from_FIXED (points[i + 1].x),
-						_cairo_fixed_from_FIXED (points[i + 1].y),
-						_cairo_fixed_from_FIXED (points[i + 2].x),
-						_cairo_fixed_from_FIXED (points[i + 2].y));
+		    cairo_fixed_t x1, y1, x2, y2;
+                    _cairo_win32_transform_FIXED_to_fixed (&transform,
+                                                           points[i].x,
+                                                           points[i].y,
+                                                           &x, &y);
+                    _cairo_win32_transform_FIXED_to_fixed (&transform,
+                                                           points[i + 1].x,
+                                                           points[i + 1].y,
+                                                           &x1, &y1);
+                    _cairo_win32_transform_FIXED_to_fixed (&transform,
+                                                           points[i + 2].x,
+                                                           points[i + 2].y,
+                                                           &x2, &y2);
+		    _cairo_path_fixed_curve_to (path, x, y, x1, y1, x2, y2);
 		}
 		break;
 	    }
@@ -1419,6 +1495,8 @@ _cairo_win32_font_face_scaled_font_create (void			*abstract_face,
 {
     cairo_win32_font_face_t *font_face = abstract_face;
 
+    _cairo_win32_initialize ();
+
     *font = _win32_scaled_font_create (&font_face->logfont,
 				       font_face->hfont,
 				       &font_face->base,
@@ -1456,6 +1534,8 @@ cairo_win32_font_face_create_for_logfontw (LOGFONTW *logfont)
 {
     cairo_win32_font_face_t *font_face;
 
+    _cairo_win32_initialize ();
+
     font_face = malloc (sizeof (cairo_win32_font_face_t));
     if (!font_face) {
 	_cairo_error (CAIRO_STATUS_NO_MEMORY);
@@ -1489,6 +1569,8 @@ cairo_win32_font_face_create_for_hfont (HFONT font)
 {
     cairo_win32_font_face_t *font_face;
 
+    _cairo_win32_initialize ();
+
     font_face = malloc (sizeof (cairo_win32_font_face_t));
     if (!font_face) {
 	_cairo_error (CAIRO_STATUS_NO_MEMORY);
@@ -1593,3 +1675,19 @@ cairo_win32_scaled_font_get_metrics_factor (cairo_scaled_font_t *scaled_font)
 {
     return 1. / ((cairo_win32_scaled_font_t *)scaled_font)->logical_scale;
 }
+
+void
+cairo_win32_scaled_font_get_logical_to_device (cairo_scaled_font_t *scaled_font,
+					       cairo_matrix_t *logical_to_device)
+{
+    cairo_win32_scaled_font_t *win_font = (cairo_win32_scaled_font_t *)scaled_font;
+    *logical_to_device = win_font->logical_to_device;
+}
+
+void
+cairo_win32_scaled_font_get_device_to_logical (cairo_scaled_font_t *scaled_font,
+					       cairo_matrix_t *device_to_logical)
+{
+    cairo_win32_scaled_font_t *win_font = (cairo_win32_scaled_font_t *)scaled_font;
+    *device_to_logical = win_font->device_to_logical;
+}
