@@ -38,6 +38,7 @@
 
 #include "cairo-test.h"
 
+#include "pdiff.h"
 #include "buffer-diff.h"
 #include "xmalloc.h"
 
@@ -64,9 +65,7 @@ buffer_diff_core (unsigned char *_buf_a,
 		  unsigned char *_buf_diff,
 		  int		width,
 		  int		height,
-		  int		stride_a,
-		  int		stride_b,
-		  int		stride_diff,
+		  int		stride,
 		  pixman_bits_t mask,
 		  buffer_diff_result_t *result_ret)
 {
@@ -77,14 +76,12 @@ buffer_diff_core (unsigned char *_buf_a,
     pixman_bits_t *buf_b = (pixman_bits_t*)_buf_b;
     pixman_bits_t *buf_diff = (pixman_bits_t*)_buf_diff;
 
-    stride_a /= sizeof(pixman_bits_t);
-    stride_b /= sizeof(pixman_bits_t);
-    stride_diff /= sizeof(pixman_bits_t);
+    stride /= sizeof(pixman_bits_t);
     for (y = 0; y < height; y++)
     {
-	row_a = buf_a + y * stride_a;
-	row_b = buf_b + y * stride_b;
-	row = buf_diff + y * stride_diff;
+	row_a = buf_a + y * stride;
+	row_b = buf_b + y * stride;
+	row = buf_diff + y * stride;
 	for (x = 0; x < width; x++)
 	{
 	    /* check if the pixels are the same */
@@ -121,19 +118,48 @@ buffer_diff_core (unsigned char *_buf_a,
 }
 
 void
-buffer_diff (unsigned char *buf_a,
-	     unsigned char *buf_b,
-	     unsigned char *buf_diff,
-	     int	   width,
-	     int	   height,
-	     int	   stride_a,
-	     int	   stride_b,
-	     int	   stride_diff,
-	     buffer_diff_result_t *result)
+compare_surfaces (cairo_surface_t	*surface_a,
+		  cairo_surface_t	*surface_b,
+		  cairo_surface_t	*surface_diff,
+		  buffer_diff_result_t	*result)
 {
-    buffer_diff_core(buf_a, buf_b, buf_diff,
-		     width, height, stride_a, stride_b, stride_diff, 0xffffffff,
-		     result);
+    /* These default values were taken straight from the
+     * perceptualdiff program. We'll probably want to tune these as
+     * necessary. */
+    double gamma = 2.2;
+    double luminance = 100.0;
+    double field_of_view = 45.0;
+    int discernible_pixels_changed;
+
+    /* First, we run cairo's old buffer_diff algorithm which looks for
+     * pixel-perfect images, (we do this first since the test suite
+     * runs about 3x slower if we run pdiff_compare first).
+     */
+    buffer_diff_core (cairo_image_surface_get_data (surface_a),
+		      cairo_image_surface_get_data (surface_b),
+		      cairo_image_surface_get_data (surface_diff),
+		      cairo_image_surface_get_width (surface_a),
+		      cairo_image_surface_get_height (surface_a),
+		      cairo_image_surface_get_stride (surface_a),
+		      0xffffffff,
+		      result);
+    if (result->pixels_changed == 0)
+	return;
+
+    cairo_test_log ("%d pixels differ (with maximum difference of %d) from reference image\n",
+		    result->pixels_changed, result->max_diff);
+
+    /* Then, if there are any different pixels, we give the pdiff code
+     * a crack at the images. If it decides that there are no visually
+     * discernible differences in any pixels, then we accept this
+     * result as good enough. */
+    discernible_pixels_changed = pdiff_compare (surface_a, surface_b,
+						gamma, luminance, field_of_view);
+    if (discernible_pixels_changed == 0) {
+	result->pixels_changed = 0;
+	cairo_test_log ("But perceptual diff finds no visually discernible difference.\n"
+			"Accepting result.\n");
+    }
 }
 
 void
@@ -142,13 +168,11 @@ buffer_diff_noalpha (unsigned char *buf_a,
 		     unsigned char *buf_diff,
 		     int	   width,
 		     int	   height,
-		     int	   stride_a,
-		     int	   stride_b,
-		     int	   stride_diff,
+		     int	   stride,
 		     buffer_diff_result_t *result)
 {
     buffer_diff_core(buf_a, buf_b, buf_diff,
-		     width, height, stride_a, stride_b, stride_diff, 0x00ffffff,
+		     width, height, stride, 0x00ffffff,
 		     result);
 }
 
@@ -195,6 +219,35 @@ flatten_surface (cairo_surface_t **surface, int x, int y)
     *surface = flat;
 }
 
+/* Given an image surface, create a new surface that has the same
+ * contents as the sub-surface with its origin at x,y.
+ *
+ * The original surface will be destroyed.
+ */
+static void
+extract_sub_surface (cairo_surface_t **surface, int x, int y)
+{
+    cairo_surface_t *sub;
+    cairo_t *cr;
+
+    sub = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
+				      cairo_image_surface_get_width (*surface) - x,
+				      cairo_image_surface_get_height (*surface) - y);
+
+    /* We don't use a device offset like flatten_surface. That's not
+     * for any important reason, (the results should be
+     * identical). This style just seemed more natural to me this
+     * time, so I'm leaving both here so I can look at both to see
+     * which I like better. */
+    cr = cairo_create (sub);
+    cairo_set_source_surface (cr, *surface, -x, -y);
+    cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
+    cairo_paint (cr);
+
+    cairo_surface_destroy (*surface);
+    *surface = sub;
+}
+
 /* Image comparison code courtesy of Richard Worth <richard@theworths.org>
  * Returns number of pixels changed, (or -1 on error).
  * Also saves a "diff" image intended to visually show where the
@@ -221,27 +274,54 @@ image_diff_core (const char *filename_a,
 		 buffer_diff_result_t *result,
 		 cairo_bool_t	flatten)
 {
+    cairo_status_t status;
     unsigned int width_a, height_a, stride_a;
     unsigned int width_b, height_b, stride_b;
     cairo_surface_t *surface_a, *surface_b, *surface_diff;
 
     surface_a = cairo_image_surface_create_from_png (filename_a);
-    if (cairo_surface_status (surface_a))
-	return cairo_surface_status (surface_a);
-
-    surface_b = cairo_image_surface_create_from_png (filename_b);
-    if (cairo_surface_status (surface_b)) {
-	cairo_surface_destroy (surface_a);
-	return cairo_surface_status (surface_b);
+    status = cairo_surface_status (surface_a);
+    if (status) {
+	cairo_test_log ("Error: Failed to create surface from %s: %s\n",
+			filename_a, cairo_status_to_string (status));
+	return status;
     }
 
-    width_a = cairo_image_surface_get_width (surface_a) - ax;
-    height_a = cairo_image_surface_get_height (surface_a) - ay;
-    width_b = cairo_image_surface_get_width (surface_b) - bx;
-    height_b = cairo_image_surface_get_height (surface_b) - by;
+    surface_b = cairo_image_surface_create_from_png (filename_b);
+    status = cairo_surface_status (surface_b);
+    if (status) {
+	cairo_test_log ("Error: Failed to create surface from %s: %s\n",
+			filename_b, cairo_status_to_string (status));
+	cairo_surface_destroy (surface_a);
+	return status;
+    }
+
+    if (flatten) {
+	flatten_surface (&surface_a, ax, ay);
+	flatten_surface (&surface_b, bx, by);
+	ax = ay = bx = by = 0;
+    }
+
+    if (ax || ay) {
+	extract_sub_surface (&surface_a, ax, ay);
+	ax = ay = 0;
+    }
+
+    if (bx || by) {
+	extract_sub_surface (&surface_b, bx, by);
+	bx = by = 0;
+    }
+
+    width_a = cairo_image_surface_get_width (surface_a);
+    height_a = cairo_image_surface_get_height (surface_a);
+    stride_a = cairo_image_surface_get_stride (surface_a);
+    width_b = cairo_image_surface_get_width (surface_b);
+    height_b = cairo_image_surface_get_height (surface_b);
+    stride_b = cairo_image_surface_get_stride (surface_b);
 
     if (width_a  != width_b  ||
-	height_a != height_b)
+	height_a != height_b ||
+	stride_a != stride_b)
     {
 	cairo_test_log ("Error: Image size mismatch: (%dx%d) vs. (%dx%d)\n"
 			"       for %s vs. %s\n",
@@ -253,27 +333,10 @@ image_diff_core (const char *filename_a,
 	return CAIRO_STATUS_SURFACE_TYPE_MISMATCH;
     }
 
-    if (flatten) {
-	flatten_surface (&surface_a, ax, ay);
-	flatten_surface (&surface_b, bx, by);
-	ax = ay = bx = by = 0;
-    }
-
     surface_diff = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
 					       width_a, height_a);
 
-    stride_a = cairo_image_surface_get_stride (surface_a);
-    stride_b = cairo_image_surface_get_stride (surface_b);
-
-    buffer_diff (cairo_image_surface_get_data (surface_a)
-		 + (ay * stride_a) + ax * 4,
-		 cairo_image_surface_get_data (surface_b)
-		 + (by * stride_b) + by * 4,
-		 cairo_image_surface_get_data (surface_diff),
-		 width_a, height_a,
-		 stride_a, stride_b,
-		 cairo_image_surface_get_stride (surface_diff),
-		 result);
+    compare_surfaces (surface_a, surface_b, surface_diff, result);
 
     if (result->pixels_changed) {
 	FILE *png_file;
