@@ -28,9 +28,9 @@
 #include "cairo-test.h"
 #include <stdio.h>
 
-#define WIDTH 64
-#define HEIGHT 64
-#define PAD 10
+#define WIDTH 16
+#define HEIGHT 16
+#define PAD 2
 
 const char	png_filename[]	= "romedalen.png";
 
@@ -126,8 +126,8 @@ draw_glyphs (cairo_t *cr, int x, int y)
 
     cairo_text_extents (cr, "FG", &extents);
     cairo_move_to (cr,
-		   x + (WIDTH - extents.width) / 2 - extents.x_bearing,
-		   y + (HEIGHT - extents.height) / 2 - extents.y_bearing);
+		   x + floor ((WIDTH - extents.width) / 2 + 0.5) - extents.x_bearing,
+		   y + floor ((HEIGHT - extents.height) / 2 + 0.5) - extents.y_bearing);
     cairo_show_text (cr, "FG");
 }
 
@@ -228,14 +228,14 @@ draw (cairo_t *cr, int width, int height)
 	    pattern_funcs[i] (cr, x, y);
 	    draw_funcs[j] (cr, x, y);
 	    if (cairo_status (cr))
-		cairo_test_log ("%d %d HERE!\n", i, j);
+		cairo_test_log ("%d %d HERE!\n", (int)i, (int)j);
 
 	    cairo_restore (cr);
 	}
     }
 
     if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
-	cairo_test_log ("%d %d .HERE!\n", i, j);
+	cairo_test_log ("%d %d .HERE!\n", (int)i, (int)j);
 
     return CAIRO_TEST_SUCCESS;
 }
